/*
  focus.c - Blue-ray Pickup laser focus module
 Part of TwinTeeh Firmware
 
 Copyright (c) 2015 V.Ayllon.  All rights reserved.
 
 TwinTeeth Firmware is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeeth Firmware, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeeth Firmware so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
#include "TwinTeeth.h"
#include "laser_control.h"
#include "focus.h"
#include "planner.h"
#include "stepper.h"

typedef struct focus_point {
  float x;
  float y;
  uint8_t focus_val;
} 
FOCUS_POINT;

FOCUS_POINT fpoints[NUM_FOCUS_POINTS];

int fe_max_val = MIN_FE_VALUE;	        // Is OK
int fe_min_val = MAX_FE_VALUE;	        // Is OK

volatile int fe_servo_pos = 0;
volatile int fe_state = FE_WAIT_ST;
volatile int fe_amplitud = 0;
volatile int fe_focus_pos = 0;
uint8_t focus_diode = LASER_BLUE;
float focus_offset = 0.;

extern void prepare_move(bool rapid, raster_t* raster_b);
extern float destination[NUM_AXIS];
extern float next_feedrate;

/**
 ** set_focus_diode() - Sets the laser diode use for focus
 **/
void set_focus_diode(uint8_t d){
  focus_diode = d;
}


/**
 ** init_focus() - Init the focus variables
 **/
void init_focus(float foffset, float width, float height){
  focus_offset = foffset;
  fpoints[0].x = -width/2.0 + focus_offset;
  fpoints[0].y = height/2.0 - focus_offset;
  fpoints[0].focus_val = FOCUS_ERROR;
  fpoints[1].x = fpoints[0].x;
  fpoints[1].y = -height/2.0 + focus_offset;
  fpoints[1].focus_val = FOCUS_ERROR;
  fpoints[2].x = width/2.0 - focus_offset;
  fpoints[2].y = fpoints[0].y;
  fpoints[2].focus_val = FOCUS_ERROR;
  fpoints[3].x = fpoints[2].x;
  fpoints[3].y = fpoints[1].y;
  fpoints[3].focus_val = FOCUS_ERROR;
}


/**
 ** get_focus_points() - Move printer to each PCB corners, focus the laser there and obtain the values
 ** Invoque determines if is the first = 0 or other invocations > 0
 **/
uint8_t get_focus_points(){


  st_synchronize();
  //
  // Check focus val if it is zero
  // it means we need to focus that point
  // Move to focus points and take focus info there
  //
  for (uint8_t i = 0; i < NUM_FOCUS_POINTS; i++){
    if (fpoints[i].focus_val == FOCUS_ERROR){
      destination[X_AXIS] = fpoints[i].x;
      destination[Y_AXIS] = fpoints[i].y;
      destination[Z_AXIS] = LASER_Z_FOCUS_POS;
      next_feedrate = 0;
      prepare_move(true,NULL);
      SERIAL_ECHOPGM(" x:");
      SERIAL_ECHO(fpoints[i].x);
      SERIAL_ECHOPGM(" y:");
      SERIAL_ECHOLN(fpoints[i].y);
      // Now focus there
      finishAndDisableSteppers(); // Disable motors to avoid noise
      //
      // Focus the laser
      //
      if ((fpoints[i].focus_val = auto_focus()) == FOCUS_ERROR){
        return FOCUS_ERROR;
      }
    }
  }
  // OK we have now the four point focus values
  return FOCUS_OK;
}

/**
 ** auto_focus() - Focus the laser
 **/
uint8_t auto_focus(){
  uint8_t focus;
  uint8_t focus_sample[NUM_FOCUS_SAMPLES];

  //
  // Get focus position
  // Try some times and get average focus
  //
  uint8_t i = 0;

  for (int n = 0; n < NUM_FOCUS_SAMPLES; n++){
    if (get_focus_position() != FOCUS_ERROR){
      if (fe_focus_pos > MIN_AMP_THRESHOLD){
        SERIAL_ECHOLN(fe_focus_pos);
        focus_sample[i++] = (uint8_t)fe_focus_pos;
      }
      delay_ms(100);
    } 
    else {
      return FOCUS_ERROR;
    }
  } 

  if (i < NUM_FOCUS_SAMPLES/2)
    return FOCUS_ERROR;
  //
  // We have some Focus positions
  // Calc mode, focus the laser & restore
  // previous laser power
  focus = calc_mode(focus_sample,NUM_FOCUS_SAMPLES);

  if (int(focus) < MIN_FOCUS_VALUE){
    SERIAL_ECHOLNPGM("Error: Focus point too close. Set zero Z axis down.");
    return FOCUS_ERROR;
  }
  // Move servo to focus position
  move_focus_servo(focus); 
  SERIAL_ECHOPGM("Focus point:");
  SERIAL_ECHOLN((int)focus);
  return focus;
}

/**
 ** focus_laser() - Try to focus the laser
 **/
uint8_t get_focus_position(){

  uint8_t last_lpower = get_laser_power();

  for (int i = 0; i < MAX_FOCUS_SWEEP; i++){
    if (focus_diode == LASER_BLUE)
      set_laser_power(BLUE_LASER_FOCUS_POWER);
    else
      set_laser_power(IR_LASER_FOCUS_POWER);

    turn_laser(focus_diode);


    move_focus_servo(FSERVO_MAX);
    delay_us(1200);


    start_fe_sampler();
    // Set state
    set_state(FE_AMPLITUD_ST);
    // Init variables
    fe_max_val = MIN_FE_VALUE;
    fe_min_val = MAX_FE_VALUE;
    fe_amplitud = 0;
    //
    // move servo from Up to Down
    //
    for (fe_servo_pos = FSERVO_MAX; fe_servo_pos > FSERVO_MIN; fe_servo_pos--){
      move_focus_servo(fe_servo_pos);
    }
    //
    // Set state to wait, while calculate amplitud
    //
    set_state(FE_WAIT_ST);
    //
    // Get Amplitud of S-Curve
    // 
    fe_amplitud = (fe_max_val-fe_min_val);
    //
    // If Amplitud has a miminum value, try to focus
    //SERIAL_ECHOPGM(" fe_amplitud: ");
    //SERIAL_ECHO((int)fe_amplitud);
    //SERIAL_ECHOPGM(" - ");
    //SERIAL_ECHO((int)fe_min_val);
    //SERIAL_ECHOPGM(" ");
    //SERIAL_ECHOLN((int)fe_max_val);

    if (fe_amplitud > MIN_AMP_THRESHOLD){
      //
      // Try to focus. Move servo Down to UP
      // and monitor FE value. When fe_state == FE_FOCUSED_ST
      // the laser is focused.

      set_state(FE_FOCUS_ST);

      while (++fe_servo_pos < FSERVO_MAX && fe_state != FE_ERROR_ST){
        move_focus_servo(fe_servo_pos);
        cli();
        if (fe_state == FE_FOCUSED_ST){ // Focused!!
          // Stop sampling and return
          sei();             
          stop_fe_sampler();
          turn_laser_off();
          set_laser_power(last_lpower);
          return 0; 
        }
        sei();
      }  
    } 
    stop_fe_sampler();
    turn_laser_off();
  }
  //
  // Max attempts, restore power
  //
  set_laser_power(last_lpower);
  return FOCUS_ERROR;
}



void start_fe_sampler(){
  cli();

  //clear ADCSRA and ADCSRB registers
  ADCSRA = 0;
  ADCSRB = 0;
  // Init ADC
  ADMUX = ((1 << REFS0) | ((FE_SIGNAL_PIN-PIN_A0) & 0x07));
  ADCSRA |= (1 << ADPS1) | (1 << ADPS0); //set ADC clock with 32 prescaler- 16mHz/4=2000kHz
  ADCSRA |= (1 << ADATE); //enabble auto trigger
  ADCSRA |= (1 << ADIE); //enable interrupts when measurement complete
  ADCSRA |= (1 << ADEN); //enable ADC
  ADCSRA |= (1 << ADSC); //start ADC measurements
  // Set analog input
  DIDR0 = 0;
  DIDR0 |= (1<<(FE_SIGNAL_PIN-PIN_A0)); 
  fe_state = FE_WAIT_ST;
  sei();
}
/**
 ** stop_fe_sampler() - Stop the sampling of data
 **/
void stop_fe_sampler(){
  cli();
  ADCSRA = 0;
  ADCSRB = 0;
  fe_state = FE_WAIT_ST;
  sei();
}
/**
 ** set_state() - Sets the state of sampling
 **/
void set_state(uint8_t state){
  cli();
  fe_state = state;
  sei();
}
/**
 ** Interrupt vector
 **/
ISR(ADC_vect) {//when new ADC value ready
  static int FEVal;
  static int FEtimeout;
  ADCSRB = 0;
  ADMUX = ((1 << REFS0) | ((FE_SIGNAL_PIN-PIN_A0) & 0x07));

  // State machine
  switch (fe_state){
  case FE_WAIT_ST:
    FEtimeout=0;
    break;
  case FE_AMPLITUD_ST:
    ADCSRA |= 1<<ADSC; // Start conversion
    FEVal = ADC;
    if ( FEVal > fe_max_val)
      fe_max_val = FEVal;
    if (FEVal < fe_min_val)
      fe_min_val = FEVal;
    break;
  case FE_FOCUS_ST:
    ADCSRA |= 1<<ADSC; // Start conversion
    FEVal = ADC;
    if (abs(REF_FE_VALUE - FEVal) > fe_amplitud/FE_NOISE_DIVIDER){
      FEtimeout=0;
      fe_state = FE_SCURVE_ST;
    } 
    else if (FEtimeout++ > FOCUS_TIMEOUT){
      fe_state = FE_ERROR_ST;
    }
    break;
  case FE_SCURVE_ST:
    ADCSRA |= 1<<ADSC; // Start conversion
    FEVal = ADC;
    if (abs(REF_FE_VALUE - FEVal) <= fe_amplitud/AMPLITUD_DIVIDER){
      fe_focus_pos = fe_servo_pos;
      fe_state = FE_FOCUSED_ST;
    } 
    else if (FEtimeout++ > FOCUS_TIMEOUT){
      fe_state = FE_ERROR_ST;
    }
    break;
  case FE_FOCUSED_ST:
    break;
  case FE_ERROR_ST:
    break;
  }
}


/**
 ** calc_mode() - Calculates the mode of an array of unsigned int numbers
 **/
uint8_t calc_mode(uint8_t *value, uint8_t num)
{
  uint8_t count, old_count;
  uint8_t mode, old_mode;

  old_count = mode =  old_mode = 0;
  uint8_t i;
  for(i=0; i<num; i++) {
    mode = value[i];
    count = 1;
    uint8_t j;
    for(j = i+1; j < num; j++) 
      if(mode==value[j]) count++;
    if(count > old_count) {
      old_mode = mode;
      old_count = count;
    }
  }
  return old_mode;
}


/**
 ** interpolate() - Bilinear interpolation of focus points
 ** returns the focus len position for a x,y point
 **/
uint8_t interpolate(float x, float y)
{
  float x2minusx1 = fpoints[2].x-fpoints[0].x;
  float y2minusy1 = fpoints[1].y-fpoints[0].y;
  float x2minusx =  fpoints[2].x-x;
  float y2minusy = fpoints[1].y-y;
  float xminusx1 = x-fpoints[0].x;
  float yminusy1 = y-fpoints[0].y;

  return int((1.0/(x2minusx1*y2minusy1)) * ((fpoints[0].focus_val*x2minusx*y2minusy) + (fpoints[2].focus_val*xminusx1*y2minusy) + (fpoints[1].focus_val*x2minusx*yminusy1) + (fpoints[3].focus_val*xminusx1*yminusy1)));
}


























































